import os

VERSION: str = "0.37.0"
REPOSITORY_URL: str = f"https://pypi.org/pypi/mindgard/json"

ALGORITHMS = ['RS256']
AUTH0_DOMAIN = os.getenv("AUTH0_DOMAIN", "login.sandbox.mindgard.ai")
AUTH0_CLIENT_ID = os.getenv("AUTH0_CLIENT_ID", "U0OT7yZLJ4GEyabar11BENeQduu4MaNO")
AUTH0_AUDIENCE = os.getenv("AUTH0_AUDIENCE", "https://marketplace-orchestrator.com")

API_BASE = os.getenv("API_BASE", "https://api.sandbox.mindgard.ai/api/v1")


from argparse import Namespace
import sys
from typing import Any, Dict, Optional, Tuple

import toml
import requests

from .constants import REPOSITORY_URL, VERSION

class CliResponse():
    def __init__(self, code:int):
        self._code = code

    def code(self) -> int:
        return self._code

def print_to_stderr(*args: Any, **kwargs: Any) -> None:
    print(*args, file=sys.stderr, **kwargs)


def version_to_tuple(version: str) -> Tuple[int, ...]:
    return tuple(map(int, version.split(".")))


def api_get(url: str, access_token: str) -> requests.Response:
    res = requests.get(url, headers=standard_headers(access_token))
    res.raise_for_status()
    return res


def api_post(url: str, access_token: str, json: Dict[str, Any]) -> requests.Response:
    res = requests.post(url, headers=standard_headers(access_token), json=json)
    res.raise_for_status()
    return res
        

def is_version_outdated() -> Optional[str]:
    try:
        res = requests.get(REPOSITORY_URL)
        res.raise_for_status()
        latest_version = res.json()["info"]["version"]
        latest_version_tuple = version_to_tuple(latest_version)
        current_version_tuple = version_to_tuple(VERSION)
        return latest_version if latest_version_tuple > current_version_tuple else None
    except Exception:
        return ''
    

def standard_headers(access_token: str) -> Dict[str, str]:
    return {
        "Authorization": f"Bearer {access_token}",
        "User-Agent": f"mindgard-cli/{VERSION}",
        "X-User-Agent": f"mindgard-cli/{VERSION}"
    }
    
def parse_toml_and_args_into_final_args(config_file_path: Optional[str], args: Namespace) -> Dict[str, Any]:
    config_file = config_file_path or "mindgard.toml"
    toml_args = {}
    try:
        with open(config_file, 'r') as f:
            contents = f.read()
            toml_args = toml.loads(contents)
    except FileNotFoundError:
        if config_file_path is None:
            pass
        else:
            raise ValueError(f"Config file not found: {config_file=}. Check that the file exists on disk.")

    final_args = {k: v or toml_args.get(k) or toml_args.get(k.replace("_", "-")) for k, v in vars(args).items()}

    return final_args




import os
import time
from typing import Any, Callable, Optional, TypeVar, cast
from functools import wraps

from rich.progress import Console

import requests
from auth0.authentication.token_verifier import (AsymmetricSignatureVerifier, # type: ignore
                                                 TokenVerifier)

from .constants import AUTH0_AUDIENCE, AUTH0_CLIENT_ID, AUTH0_DOMAIN
from .utils import CliResponse, print_to_stderr



def get_config_directory() -> str:
    config_dir = os.environ.get('MINDGARD_CONFIG_DIR')
    return config_dir or os.path.join(os.path.expanduser('~'), '.mindgard')


def get_token_file() -> str:
    return os.path.join(get_config_directory(), 'token.txt')


def clear_token() -> None:
    if os.path.exists(get_token_file()):
        os.remove(get_token_file())


# TODO: test
def validate_id_token(id_token: str) -> None:
    """
    Verify the token and its precedence. Raises if the token is invalid.

    :param id_token: string
    """
    jwks_url = 'https://{}/.well-known/jwks.json'.format(AUTH0_DOMAIN)
    issuer = 'https://{}/'.format(AUTH0_DOMAIN)
    sv = AsymmetricSignatureVerifier(jwks_url)
    tv = TokenVerifier(signature_verifier=sv, issuer=issuer, audience=AUTH0_CLIENT_ID)
    tv.verify(id_token)


def load_access_token() -> Optional[str]:
    """
    Reads refresh token from file or environment variable and returns a new access token
    Returns None if no refresh token is found
    """
    refresh_token = os.environ.get('MINDGARD_API_KEY')
    if not refresh_token and os.path.exists(get_token_file()):
        with open(get_token_file(), 'r') as f:
            refresh_token = f.read()

    if refresh_token:
        access_token = requests.post(
            'https://{}/oauth/token'.format(AUTH0_DOMAIN),
            data={
                'grant_type': 'refresh_token',
                'client_id': AUTH0_CLIENT_ID,
                'audience': AUTH0_AUDIENCE,
                'refresh_token': refresh_token
            }
        ).json().get('access_token')
        return cast(str, access_token)
    return None
    

def login() -> None:
    """
    Runs the device authorization flow and stores the user token in memory
    """

    print("Welcome to Mindgard! Let\'s get you authenticated...")
    print("\033[1;37mNote: Mindgard is an AI security testing tool that will run red-team attacks to assess the risk of the AI systems you are testing.")
    print("Only use Mindgard with systems you have authorization to test in this manner.\033[0;0m\n")
    print("By continuing you acknowledge this and the terms of service.\n")

    device_code_payload = {
        'client_id': AUTH0_CLIENT_ID,
        'scope': 'openid profile email offline_access',
        'audience': AUTH0_AUDIENCE
    }
    device_code_response = requests.post('https://{}/oauth/device/code'.format(AUTH0_DOMAIN), data=device_code_payload)

    if device_code_response.status_code != 200:
        print_to_stderr('Error generating login url. Please try again. Contact Mindgard support if the issue persists.')
        raise ValueError(f"Login service API response: {device_code_response.json()}")

    device_code_data = device_code_response.json()
    print('1. On your computer or mobile device navigate to: ', device_code_data['verification_uri_complete'])
    print('2. Confirm that you see the following code: ', device_code_data['user_code'])
    print('3. Register/log in using the web UI\n')


    # New code 👇
    token_payload = {
        'grant_type': 'urn:ietf:params:oauth:grant-type:device_code',
        'device_code': device_code_data['device_code'],
        'client_id': AUTH0_CLIENT_ID,
        'audience': AUTH0_AUDIENCE
    }

    console = Console()
    with console.status("[bold green]Waiting for auth to complete...") as status:
        authenticated = False
        while not authenticated:
            # print('Checking if the login flow has been completed...    no pressure!')
            token_response = requests.post('https://{}/oauth/token'.format(AUTH0_DOMAIN), data=token_payload)

            token_data = token_response.json()
            if token_response.status_code == 200:
                validate_id_token(token_data['id_token'])
                print('Authenticated!')
                os.makedirs(get_config_directory(), exist_ok=True)
                with open(get_token_file(), 'w') as f:
                    f.write(token_data['refresh_token'])
                authenticated = True
            elif token_data['error'] not in ('authorization_pending', 'slow_down'):
                error = token_data.get('error_description', 'Error authenticating the user. Please wait 30s and try again.')
                raise ValueError(error)
            else:
                time.sleep(device_code_data['interval'])



def logout() -> None:
    """
    Removes the user token
    """
    print(f'Clearing credentials stored at {get_token_file()}.')
    clear_token()
    print('Logged out!')
    
    
T = TypeVar('T')
# TODO: improve typing definitions here
def require_auth(func: Callable[..., CliResponse]) -> Callable[..., CliResponse]:
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> CliResponse:
        access_token = load_access_token()
        if not access_token:
            print_to_stderr("\033[1;37mRun `mindgard login`\033[0;0m to authenticate.")
            return CliResponse(2)
        try:
            return func(*args, **kwargs, access_token = access_token)
        except requests.HTTPError as e:
            if "Unauthorized" in str(e):
                print_to_stderr("Access token is invalid. Please re-authenticate using `mindgard login`")
                clear_token()
                return CliResponse(2)
            else:
                raise e
    return wrapper





import json
from typing import Optional
from rich.console import Console, Group

from rich.panel import Panel
from rich.align import Align

from .utils import CliResponse

from .auth import require_auth
from .api_service import ApiService


class ListTestsCommand():
    
    def __init__(self, api_service: ApiService) -> None:
        self._api = api_service

    @require_auth
    def run(self, access_token:str, json_format:bool, test_id:Optional[str]) -> CliResponse:
        """
        Run the command.

        Returns int of exit code
        """

        console = Console()
        
        if test_id is None:
            tests_res = self._api.get_tests(access_token)
        else:
            tests_res = [self._api.get_test(access_token, test_id=test_id)]

        if json_format is True:
            print(json.dumps(tests_res))
        else:
            for test in tests_res:

                score = test["risk"]
                model_name = test["mindgardModelName"]
                url = test["url"]
                
                risk_text = 'Minimal'
                if score >= 90:
                    risk_text = 'Critical'
                elif score >= 66:
                    risk_text = 'High'
                elif score >= 40:
                    risk_text = 'Medium'
                elif score >= 20:
                    risk_text = 'Low'
                
                def bold(input:str) -> str:
                    return "[b]" + input + "[/b]"
                
                layout = Group(
                    Align.center(bold(model_name)),
                    "Total tests: " + bold(str(138)),
                    "Last test:", 
                    "  model name: " +  bold(model_name),
                    "  tested at: " + bold(test["createdAt"]),
                    "  threat level: " + bold(risk_text),
                    "  risk score: " + bold(str(score)),
                    f"  details: {url}"
                )
                panel = Panel(layout)
                
                
                console.print(panel)

        return CliResponse(0)







import json
from typing import Dict, Any
from time import sleep
from concurrent.futures import ThreadPoolExecutor
from rich.console import Console
from rich.table import Table
from rich.live import Live
from rich.progress import Progress, SpinnerColumn, TaskID

from .utils import CliResponse

from .auth import require_auth
from .api_service import ApiService

TEST_POLL_INTERVAL = 5
class RunTestCommand(): # called 'RunTest..' to avoid pytest thinking it's a test
    """
    Command to execute a single test and watch the results
    """
    
    def __init__(self, api_service: ApiService, poll_interval:float = TEST_POLL_INTERVAL) -> None:
        self._api = api_service
        self._poll_interval = poll_interval # poll interval is expose to speed up tests

    def submit_test_progress(self, progress:Progress, access_token:str, model_name:str) -> Dict[str,Any]:
        with progress:
            with ThreadPoolExecutor() as pool:
                task_id = progress.add_task("submitting test", start=True)

                future = pool.submit(self.submit_test_fetching_initial, access_token, model_name)
                while not future.done():
                    progress.update(task_id, refresh=True)
                    sleep(0.1)
                progress.update(task_id, completed=100)
                return future.result()

    def submit_test_fetching_initial(self, access_token:str, model_name:str) -> Dict[str, Any]:
        initial_resp = self._api.submit_test(access_token, model_name)
        return self._api.get_test(access_token, initial_resp["id"])
        
    def run_inner(self, access_token:str, model_name:str, json_format:bool, risk_threshold:int) -> CliResponse:
        if json_format:
            return self.run_json(access_token=access_token, model_name=model_name, risk_threshold=risk_threshold)

        submit_progress = Progress(
            "{task.description}",
            SpinnerColumn(
                finished_text=r"\[done]"
            ),
            auto_refresh=True
        )
        
        test_res: Dict[str, Any]
        with submit_progress:
            test_res = self.submit_test_progress(submit_progress, access_token=access_token, model_name=model_name)

        attacks = test_res["attacks"]
        test_id = test_res["id"]
        attack_count = len(attacks)

        overall_progress = Progress()
        overall_task = overall_progress.add_task("overall", total=attack_count)

        attacks_progress = Progress(
            "{task.description}",
            SpinnerColumn(
                finished_text=r"\[done]"
            ),
        )
        attacks_task_map: Dict[str, TaskID] = {}
        for attack in attacks:
            attacks_task_map[attack["id"]] = attacks_progress.add_task(f"attack {attack['attack']}", total=1)
        

        progress_table = Table.grid(expand=True)
        progress_table.add_row(
            overall_progress
        )
        progress_table.add_row(
            attacks_progress
        )

        with Live(progress_table, refresh_per_second=10):
            while not overall_progress.finished:
                sleep(self._poll_interval)
                test_res = self._api.get_test(access_token, test_id=test_id)

                for attack_res in test_res["attacks"]:
                    task_id = attacks_task_map[attack_res["id"]]
                    if attack_res["state"] == 2:
                        attacks_progress.update(task_id, completed=1)
                        
                completed = sum(task.completed for task in attacks_progress.tasks)
                overall_progress.update(overall_task, completed=completed)
        

        table = Table(title=f"Results - https://sandbox.mindgard.ai/r/test/{test_id}", width=80)
        table.add_column("Pass", style="cyan")
        table.add_column("Name", style="magenta")
        table.add_column("Risk", justify="right", style="green")

        for attack in test_res["attacks"]:
            risk = attack["risk"]
            emoji = "❌‍" if risk > risk_threshold else "✅️"
            table.add_row(emoji, attack["attack"], str(risk))


        console = Console()
        console.print(table)

        return CliResponse(self.calculate_exit_code(test_res=test_res,risk_threshold=risk_threshold))

    @require_auth
    def run(self, access_token:str, model_name:str, json_format:bool, risk_threshold:int) -> CliResponse:
        """
        Run the command.

        Returns int of exit code
        """
        return self.run_inner(access_token=access_token, json_format=json_format, model_name=model_name, risk_threshold=risk_threshold)


    def run_json(self, access_token:str, model_name:str, risk_threshold:int) -> CliResponse:
        test_res = self.submit_test_fetching_initial(access_token=access_token, model_name=model_name)
        test_id = test_res["id"]
        while test_res["hasFinished"] is False:
            sleep(self._poll_interval)
            test_res = self._api.get_test(access_token, test_id=test_id)

        print(json.dumps(test_res))
        return CliResponse(self.calculate_exit_code(test_res=test_res,risk_threshold=risk_threshold))


    def calculate_exit_code(self, test_res:Dict[str, Any], risk_threshold:int) -> int:
        if test_res["risk"] > risk_threshold:
            return 1
        else:
            return 0




import json
import os
from typing import List, Dict, Any
from .utils import api_get, api_post
from .constants import API_BASE

class ApiService():
    def get_tests(self, access_token: str) ->  List[Dict[str, Any]]:
        url = f"{API_BASE}/assessments?ungrouped=true"

        res = api_get(url, access_token)
        
        data: List[Dict[str, Any]] = res.json()

        # augment the output with the shortened url
        for test in data:
            test_id = test["id"]
            test["url"] = f"https://sandbox.mindgard.ai/r/test/{test_id}"
            for attack in test["attacks"]:
                attack_id = attack["id"]
                attack["url"] = f"https://sandbox.mindgard.ai/r/attack/{attack_id}"

        return data

    def get_test(self, access_token: str, test_id:str) -> Dict[str, Any]:
        url = f"{API_BASE}/assessments/{test_id}"
        res = api_get(url, access_token)
        data: Dict[str, Any] = res.json()

        data["url"] = f"https://sandbox.mindgard.ai/r/test/{test_id}"
        for attack in data["attacks"]:
            attack_id = attack["id"]
            attack["url"] = f"https://sandbox.mindgard.ai/r/attack/{attack_id}"

        return data

    def submit_test(self, access_token: str, target_name:str) -> Dict[str, Any]:
        url = f"{API_BASE}/assessments"
        post_body = {"mindgardModelName": target_name}
        res = api_post(url, access_token, json=post_body)
        data: Dict[str, Any] = res.json()
        return data
    
    def get_orchestrator_websocket_connection_string(self, access_token: str, payload: Dict[str, Any]) -> Dict[str, Any]:
        url = f"{API_BASE}/tests/cli_init"

        # Extra configuration only currently available to Mindgardians
        extra_config = os.environ.get('MINDGARD_EXTRA_CONFIG', None)
        if extra_config:
            payload["extraConfig"] = json.loads(extra_config)
        pack = os.environ.get('ATTACK_PACK', "sandbox")
        payload["attackPack"] = pack
        # End of Mindgard-only configuration

        res = api_post(url, access_token, json=payload)
        data: Dict[str, Any] = res.json()
        return data


import json
import logging
from typing import Dict, Any, Callable, Literal, Optional, List
from time import sleep
from concurrent.futures import ThreadPoolExecutor
from rich.console import Console
from rich.table import Table
from rich.live import Live
from rich.progress import Progress, SpinnerColumn, TaskID, TextColumn

# Networking
from azure.messaging.webpubsubclient import WebPubSubClient, WebPubSubClientCredential
from azure.messaging.webpubsubclient.models import OnGroupDataMessageArgs

# Exceptions
from .exceptions import (
    MGException,
    ServiceUnavailable,
    InternalServerError,
    RateLimitOrInsufficientCredits,
    FailedDependency,
    UnprocessableEntity,
    Timeout,
    NotFound,
    BadRequest,
    Forbidden,
    Unauthorized,
    Uncontactable,
    HTTPBaseError,
    EmptyResponse
)

# Types
from typing import Type

import time

from .wrappers import ModelWrapper, ContextManager

from .utils import CliResponse

from .auth import require_auth
from .api_service import ApiService

TEST_POLL_INTERVAL = 5
ErrorCode = Literal["CouldNotContact", "ContentPolicy", "CLIError", "NotImplemented", "NoResponse", "RateLimited", "NetworkIssue", "MaxContextLength"]


exceptions_to_cli_status_codes: Dict[Type[Exception], ErrorCode] = {
    Uncontactable: "CouldNotContact",
    BadRequest: "NoResponse", # not sure about this, we don't handle 400 atm
    Unauthorized: "NoResponse",
    Forbidden: "NoResponse",
    NotFound: "NoResponse",
    Timeout: "NoResponse",
    UnprocessableEntity: "NoResponse", # this is currently being handled as a rate limit issue for some reason
    FailedDependency: "NoResponse",
    RateLimitOrInsufficientCredits: "RateLimited",
    InternalServerError: "NoResponse",
    ServiceUnavailable: "NoResponse",
    NotImplementedError: "NotImplemented",
    EmptyResponse: "NoResponse"
}


def handle_exception_callback(exception: Exception, handle_visual_exception_callback: Optional[Callable[[str], None]]) -> ErrorCode:
    # TODO - come take a look at this
    error_code: ErrorCode = exceptions_to_cli_status_codes.get(type(exception), "CLIError") # type: ignore
    callback_text = str(exception)

    if handle_visual_exception_callback:
        handle_visual_exception_callback(callback_text)

    logging.debug(exception)
    return error_code


class RunLLMLocalCommand:
    def __init__(
        self,
        api_service: ApiService,
        model_wrapper: ModelWrapper,
        parallelism: int,   
        poll_interval: float = TEST_POLL_INTERVAL,
    ) -> None:
        self._api = api_service
        self._poll_interval = poll_interval  # poll interval is expose to speed up tests
        self._model_wrapper = model_wrapper
        self._parallelism = parallelism
        self._context_manager = ContextManager()

    @staticmethod
    def validate_args(args:Dict[str, Any]) -> None:
        # args must include non-zero values for 
        # target: str, json_format: bool, risk_threshold: int, system_prompt: str
        missing_args: List[str]= []
        if args.get("target") == None or args.get("target") == "":
            missing_args.append("target")
        if args.get("system_prompt", None) is None:
            missing_args.append("system_prompt")
        if args['parallelism'] < 1:
            raise ValueError(f"--parallelism must be a positive integer")
        if len(missing_args) > 0:
            raise ValueError(f"Missing required arguments: {', '.join(missing_args)}")


    def submit_test_progress(
        self, progress: Progress, access_token: str, target: str, system_prompt: str, error_callback: Callable[[Exception], ErrorCode]
    ) -> Dict[str, Any]:
        with progress:
            with ThreadPoolExecutor() as pool:
                task_id = progress.add_task("Submitting test...", start=True)

                future = pool.submit(
                    self.submit_test_fetching_initial, access_token, target, system_prompt, error_callback
                )

                while not future.done():
                    progress.update(task_id, refresh=True)
                    sleep(0.1)
                progress.update(task_id, completed=100)
                return future.result()

    def submit_test_fetching_initial(
        self, access_token: str, target: str, system_prompt: str, error_callback: Callable[[Exception], ErrorCode]
    ) -> Dict[str, Any]:
        ws_token_and_group_id = (
            self._api.get_orchestrator_websocket_connection_string(
                access_token=access_token, payload={"target": target, "system_prompt": system_prompt, "parallelism": self._parallelism}
            )
        )

        url = ws_token_and_group_id.get("url", None)
        group_id = ws_token_and_group_id.get("groupId", None)

        if url is None:
            raise Exception(
                "URL from API server missing for LLM forwarding!"
            )
        if group_id is None:
            raise Exception(
                "groupId from API server missing for LLM forwarding!"
            )

        # Should fire the connect event on the orchestrator
        credentials = WebPubSubClientCredential(client_access_url_provider=url)
        ws_client = WebPubSubClient(credential=credentials)

        self.submitted_test = False
        self.submitted_test_id = ""

        # this guy is difficult to manage exceptions for as websocket messages are event-based.
        # the recv message handler also has no idea which messages are for which attack, so
        # we can't mark an attack as failed until we are complete
        def recv_message_handler(msg: OnGroupDataMessageArgs) -> None:
            if msg.data["messageType"] == "Request":
                logging.debug(f"received request {msg.data=}")
                context_id = msg.data["payload"].get("context_id", None)
                context = self._context_manager.get_context_or_none(context_id)
                content = msg.data["payload"]["prompt"]

                response: str = ""
                error_code: Optional[ErrorCode] = None

                try:
                    # would pose being explicit with __call__ so we can ctrl+f easier, not a very clear shorthand
                    response = self._model_wrapper.__call__(
                        content=content,
                        with_context=context,
                    )
                except MGException as mge:
                    error_code = error_callback(mge)
                    if error_code == "CLIError":
                        raise mge
                except Exception as e:
                    raise e
                finally:
                    # we always try to send a response
                    replyData = {
                        "correlationId": msg.data["correlationId"],
                        "messageType": "Response",
                        "status": "ok", # Deprecated but kept for compatibility
                        "payload": {
                            "response": response,
                            "error": error_code,
                        }
                    }
                    logging.debug(f"sending response {replyData=}")

                    ws_client.send_to_group("orchestrator", replyData, data_type="json") # type: ignore
            elif msg.data["messageType"] == "StartedTest": # should be something like "Submitted", upstream change required.
                self.submitted_test_id = msg.data["payload"]["testId"]
                self.submitted_test = True
            else:
                pass

        ws_client.open()

        ws_client.subscribe("group-message", recv_message_handler) # type: ignore

        payload = {
            "correlationId": "",
            "messageType": "StartTest",
            "payload": {"groupId": ws_token_and_group_id["groupId"]},
        }

        ws_client.send_to_group(group_name="orchestrator", content=payload, data_type="json") # type: ignore

        # wait 
        max_attempts = 30
        attempts_remaining = max_attempts
        while not self.submitted_test:
            time.sleep(1)
            attempts_remaining -= 1
            if attempts_remaining == 0:
                break

        # we're here if submitting was a success...

        if attempts_remaining !=0:
            return self._api.get_test(access_token, self.submitted_test_id)
        else:
            raise Exception(f"did not receive notification of test submitted within timeout ({max_attempts}s); failed to start test")

    def preflight(self, console: Console) -> bool:
        """
        Makes a single request to the LLM to validate basic connectivity before submitting
        test.

        Returns True on success, False on failure
        """
        try:
            console.print("[white]Contacting model...")
            for i in range(5):
                _ = self._model_wrapper.__call__("Hello llm, are you there?")
            return True
        except Uncontactable as cerr:
            logging.debug(cerr)
            model_api = self._model_wrapper.api_url if hasattr(self._model_wrapper, "api_url") else "<unknown>"
            console.print(f"[red]Could not connect to the model! [white](URL: {model_api}, are you sure it's correct?)")
        except HTTPBaseError as httpbe:
            logging.debug(httpbe)
            message: str = f"[red]Model pre-flight check returned {httpbe.status_code} ({httpbe.status_message})"
            console.print(message)
        except Exception as e:
            # something we've not really accounted for caught
            logging.error(e)
            raise e
        
        return False

    def run_inner(
        self, access_token: str, target: str, json_format: bool, 
        risk_threshold: int, system_prompt: str
    ) -> CliResponse:
        console = Console()

        if self.preflight(console=console) == False:
            return CliResponse(2)

        if json_format:
            return self.run_json(
                access_token=access_token, target=target, risk_threshold=risk_threshold, system_prompt=system_prompt
            )

        progress_table = Table.grid(expand=True)

        submit_progress = Progress(
            "{task.description}",
            SpinnerColumn(finished_text="[green3] Submitted!"),
            auto_refresh=True,
        )

        exceptions_progress = Progress(
            "{task.description}"
        )

        exceptions_task_table: Dict[str, TaskID] = {}
        exceptions_count_table: Dict[str, int] = {}

        def _handle_visual_exception_callback(text: str) -> None:
            if not exceptions_task_table.get(text, None):
                exceptions_count_table[text] = 0
                exceptions_task_table[text] = exceptions_progress.add_task("")

            exceptions_count_table[text] += 1
            exceptions_progress.update(
                exceptions_task_table[text],
                description=f"[dark_orange3][!!!] {text} (x{exceptions_count_table[text]})"
            )


        test_res: Dict[str, Any]
        with submit_progress:
            test_res = self.submit_test_progress(
                submit_progress, access_token=access_token, target=target, system_prompt=system_prompt,
                error_callback=lambda x: handle_exception_callback(x, _handle_visual_exception_callback)
            )

        attacks = test_res["attacks"]
        test_id = test_res["id"]
        attack_count = len(attacks)

        overall_progress = Progress()
        overall_task = overall_progress.add_task("Assessment Progress", total=attack_count)

        attacks_progress = Progress(
            "{task.description}",
            SpinnerColumn(finished_text="done"),
            TextColumn("{task.fields[status]}")
        )
        attacks_task_map: Dict[str, TaskID] = {}
        for attack in attacks:
            attacks_task_map[attack["id"]] = attacks_progress.add_task(
                f"Attack {attack['attack']}", total=1, status="[chartreuse1]queued"
            )

        progress_table.add_row(overall_progress)
        progress_table.add_row(attacks_progress)
        progress_table.add_row("")
        progress_table.add_row(exceptions_progress)


        with Live(progress_table, refresh_per_second=10):
            while not overall_progress.finished:
                sleep(self._poll_interval)
                test_res = self._api.get_test(access_token, test_id=test_id)

                for attack_res in test_res["attacks"]:
                    task_id = attacks_task_map[attack_res["id"]]
                    if attack_res["state"] == 2:
                        attacks_progress.update(task_id, completed=1, status="[chartreuse3]success")
                    elif attack_res["state"] == -1:
                        attacks_progress.update(task_id, completed=1, status="[red3]failed")
                    elif attack_res["state"] == 1:
                        attacks_progress.update(task_id, status="[orange3]running")


                completed = sum(task.completed for task in attacks_progress.tasks)
                overall_progress.update(overall_task, completed=completed)

        table = Table(
            title=f"Results - https://sandbox.mindgard.ai/r/test/{test_id}", width=80
        )
        table.add_column("Pass", style="cyan")
        table.add_column("Name", style="magenta")
        table.add_column("Risk", justify="right", style="green")

        for attack in test_res["attacks"]:
            if attack["state"] != 2:
                attack_name = attack["attack"]
                name = f"Error running '{attack_name}'"
                risk_str = "n/a"
                emoji = "❗️"
            else:
                name = attack["attack"]
                risk = attack["risk"]
                risk_str = str(risk)
                emoji = "❌‍" if risk > risk_threshold else "✅️"

            table.add_row(emoji, name, risk_str)

        console.print(table)

        return CliResponse(
            self.calculate_exit_code(test_res=test_res, risk_threshold=risk_threshold)
        )

    @require_auth
    def run(
        self, access_token: str, target: str, json_format: bool, risk_threshold: int, system_prompt: str
    ) -> CliResponse:
        """
        Run the command.

        Returns int of exit code
        """
        return self.run_inner(
            access_token=access_token,
            json_format=json_format,
            target=target,
            risk_threshold=risk_threshold,
            system_prompt=system_prompt,
        )

    def run_json(
        self, access_token: str, target: str, risk_threshold: int, system_prompt:str
    ) -> CliResponse:
        test_res = self.submit_test_fetching_initial(
            access_token=access_token, target=target, system_prompt=system_prompt, error_callback=lambda x: handle_exception_callback(x, None)
        )
        test_id = test_res["id"]
        while test_res["hasFinished"] is False:
            sleep(self._poll_interval)
            test_res = self._api.get_test(access_token, test_id=test_id)

        print(json.dumps(test_res))
        return CliResponse(
            self.calculate_exit_code(test_res=test_res, risk_threshold=risk_threshold)
        )

    def calculate_exit_code(self, test_res: Dict[str, Any], risk_threshold: int) -> int:
        if test_res["risk"] > risk_threshold:
            return 1
        else:
            return 0





import argparse
import sys
import traceback
from typing import List, cast

from .wrappers import parse_args_into_model

from .list_tests_command import ListTestsCommand
from .run_test_command import RunTestCommand
from .run_llm_local_command import RunLLMLocalCommand

from .api_service import ApiService

from .auth import login, logout
from .constants import VERSION
from .utils import is_version_outdated, print_to_stderr, parse_toml_and_args_into_final_args

import logging
from rich.logging import RichHandler
from rich.console import Console

def parse_args(args: List[str]) -> argparse.Namespace:
    log_levels = ['critical', 'fatal', 'error', 'warn', 'warning', 'info', 'debug', 'notset'] # [n.lower() for n in logging.getLevelNamesMapping().keys()]
    default_log_level = 'warn'

    parser = argparse.ArgumentParser(description='Securing AIs', prog='mindgard', usage='%(prog)s [command] [options]', epilog='Enjoy the program! :)', add_help=True)
    parser.add_argument('--version', action='version', version=f"%(prog)s {VERSION}", help='Show the current version number')
    parser.add_argument('--log-level', type=str, help='Specify the output verbosity', choices=log_levels, required=False, default=default_log_level)

    subparsers = parser.add_subparsers(dest='command', title='commands', description='Use these commands to interact with the Mindgard API')
    subparsers.add_parser('login', help='Login to the Mindgard platform')
    subparsers.add_parser('logout', help='Logout of the Mindgard platform in the CLI')

    sandbox_test_parser = subparsers.add_parser('sandbox', help='Test a mindgard example model')
    sandbox_test_parser.add_argument('target', nargs='?', type=str, choices=['cfp_faces', 'mistral'])
    sandbox_test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    sandbox_test_parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False, default=80)

    list_parser = subparsers.add_parser('list', help='List items')
    list_subparsers = list_parser.add_subparsers(dest='list_command')
    list_test_parser = list_subparsers.add_parser('tests', help='List tests')
    list_test_parser.add_argument('--json', action="store_true", help='Return json output', required=False)
    list_test_parser.add_argument('--id', type=str, help='Get the details of a specific test.', required=False)

    test_parser = subparsers.add_parser('test', help='Attack commands')
    test_parser.add_argument('target', nargs='?', type=str, help="This is your own model identifier.")
    test_parser.add_argument('--config-file', type=str, help='Path to mindgard.toml config file', default=None, required=False)
    test_parser.add_argument('--risk-threshold', type=int, help='Set a risk threshold above which the system will exit 1', required=False, default=80)
    test_parser.add_argument('--json', action="store_true", help='Output the info in JSON format.', required=False)
    test_parser.add_argument('--headers', type=str, help='The headers to use', required=False)
    test_parser.add_argument('--preset', type=str, help='The preset to use', choices=['huggingface', 'openai', 'anthropic', 'azure-openai', 'azure-aistudio', 'custom_mistral', 'tester'], required=False)
    test_parser.add_argument('--api-key', type=str, help='Specify the API key for the wrapper', required=False)
    test_parser.add_argument('--url', type=str, help='Specify the url for the wrapper', required=False)
    test_parser.add_argument('--model-name', type=str, help='Specify which model to run against (OpenAI and Anthropic)', required=False)
    test_parser.add_argument('--az-api-version', type=str, help='Specify the Azure OpenAI API version (Azure only)', required=False)
    test_parser.add_argument('--prompt', type=str, help='Specify the prompt to use', required=False)
    test_parser.add_argument('--system-prompt', type=str, help='Text file containing system prompt to use.', required=False)
    test_parser.add_argument('--selector', type=str, help='The selector to retrieve the text response from the LLM response JSON.', required=False)
    test_parser.add_argument('--request-template', type=str, help='The template to wrap the API request in.', required=False)
    test_parser.add_argument('--tokenizer', type=str, help='Choose a HuggingFace model to provide a tokeniser for prompt and chat completion templating.', required=False)
    test_parser.add_argument('--parallelism', type=int, help='The maximum number of parallel requests that can be made to the API.', required=False, default=5)

    return parser.parse_args(args)


def main() -> None:
    args = parse_args(sys.argv[1:])

    FORMAT = "%(message)s"
    logging.basicConfig(
        level=args.log_level.upper(), format=FORMAT, datefmt="[%X]", handlers=[RichHandler(console=Console(stderr=True))]
    )

    if not (sys.version_info.major == 3 and sys.version_info.minor >= 8):
        print_to_stderr("Python 3.8 or later is required to run the Mindgard CLI.")
        sys.exit(2)

    if new_version := is_version_outdated():
        print_to_stderr(f"New version available: {new_version}. Please upgrade as older versions of the CLI may not be actively maintained.")

    if args.command == 'login':
        login()
    elif args.command == 'logout':
        logout()
    elif args.command == 'list':
        if args.list_command == 'tests':
            api_service = ApiService()
            cmd = ListTestsCommand(api_service)
            res = cmd.run(json_format=bool(args.json), test_id=args.id)
            exit(res.code())
        else:
            print_to_stderr('Provide a resource to list. Eg `list tests`.')
    elif args.command == 'sandbox':
        api_service = ApiService()
        run_test_cmd = RunTestCommand(api_service)
        run_test_res = run_test_cmd.run(model_name=args.target, json_format=bool(args.json), risk_threshold=int(args.risk_threshold))
        exit(run_test_res.code())
    elif args.command == 'test':
        # load args from file mindgard.toml
        final_args = parse_toml_and_args_into_final_args(args.config_file, args)
        model_wrapper = parse_args_into_model(final_args)
        RunLLMLocalCommand.validate_args(final_args)
        api_service = ApiService()
        parallelism = int(cast(str, final_args["parallelism"]))
        llm_test_cmd = RunLLMLocalCommand(api_service=api_service, model_wrapper=model_wrapper, parallelism=parallelism)
        llm_test_res = llm_test_cmd.run(
            target=final_args["target"], 
            json_format=bool(final_args["json"]), 
            risk_threshold=int(cast(str, final_args["risk_threshold"])), 
            system_prompt=final_args["system_prompt"],
        )
        exit(llm_test_res.code())
    else:
        print_to_stderr('Which command are you looking for? See: $ mindgard --help')


if __name__ == '__main__':
    try:
        main()
    except ValueError as e:
        print_to_stderr(str(e))
        exit(2)
    except Exception:
        traceback.print_exc()
        exit(2)
